package particles;
import java.util.*;
import java.awt.Color;
import java.awt.geom.*;
import testing.*;

/* Test cases for your ParticleSystem type. Feel free to add additional test cases here
 * to exercise edge cases not covered by the provided tests.
 */
public class ParticleTests {  
    /* Utility type representing a particle that was drawn to the screen. This is used
     * by the testing code so that we can call drawParticles to store which particles
     * were drawn and where.
     */
    private static final class DrawnParticle {
        public double x, y;
        public Color color;
            
        public DrawnParticle(double x, double y, Color color) {
            this.x     = x;
            this.y     = y;
            this.color = color;
        }
        
        @Override public String toString() {
            return "(" + x + ", " + y + ", " + color + ")";
        }
        
        @Override public boolean equals(Object rhsObj) {
            if (!(rhsObj instanceof DrawnParticle)) {
                return false;
            }
            
            var rhs = (DrawnParticle) rhsObj;
            return color.equals(rhs.color) &&
                   Math.abs(x - rhs.x) <= 1e-8 &&
                   Math.abs(y - rhs.y) <= 1e-8;
        }
    }
    
    /* Utility type that acts like a ParticleGraphics and writes down all the particles
     * sent to drawParticle so that they can be inspected later.
     */
    private static class ParticleCatcher extends ParticleGraphics {        
        /* List of particles drawn thus far. */
        private List<DrawnParticle> drawn = new ArrayList<>();
        
        /* Color used for graphics operations; defaults to a funny color so
         * that it is unlikely to match existing colors.
         */
        private Color color = new Color(137, 42, 103);
        
        @Override public void setColor(Color c) {
            color = c;
        }
        
        @Override public Color getColor() {
            return color;
        }
        
        @Override public void drawParticle(double x, double y) {
            drawn.add(new DrawnParticle(x, y, color));
        }
        
        public int numDrawn() {
            return drawn.size();
        }
        
        public DrawnParticle get(int index) {
            return drawn.get(index);
        }
        
        public void reset() {
            drawn.clear();
        }
    }
    
    public static void main(String[] args) {
        new TestDriver() {
            @TestCase(milestone = 1, name = "Constructor creates an empty particle system.") public void test1() {
                var system = new ParticleSystem();
                EXPECT_EQUAL(system.head, null);
            }

            @TestCase(milestone = 1, name = "Empty system has no particles.") public void test2() {
                var system = new ParticleSystem();
                EXPECT_EQUAL(system.numParticles(), 0);
            }

            @TestCase(milestone = 1, name = "Can add a single particle.") public void test3() {
                var system = new ParticleSystem();

                Particle particle = new Particle();
                particle.x = 13.7;
                particle.y = 4.2;
                particle.color = Color.CYAN;

                system.add(particle);

                /* Should have one particle. */
                EXPECT_EQUAL(system.numParticles(), 1);

                /* Invasively check to make sure the head pointer isn't null,
                 * since it needs to hold our particle.
                 */
                EXPECT_NOT_EQUAL(system.head, null);

                /* Make sure the particle's x, y, and color are copied over. */
                EXPECT_EQUAL(system.head.particle.x, particle.x);
                EXPECT_EQUAL(system.head.particle.y, particle.y);
                EXPECT_EQUAL(system.head.particle.color, particle.color);

                /* Make sure the list is wired correctly. */
                EXPECT_EQUAL(system.head.prev, null);
                EXPECT_EQUAL(system.head.next, null);
            }

            @TestCase(milestone = 1, name = "Can add two particles.") public void test4() {
                var system = new ParticleSystem();

                /* Make two particles. Use the x and y coordinates to tell them apart. */
                Particle one = new Particle(), two = new Particle();
                one.x = 1;
                two.x = 2;

                /* Add each to the particle system. The order should be one, then two. */
                system.add(one);
                system.add(two);

                /* Make sure we see two particles. */
                EXPECT_EQUAL(system.numParticles(), 2);

                /* Make sure the list has two items in it. */
                EXPECT_NOT_EQUAL(system.head, null);
                EXPECT_NOT_EQUAL(system.head.next, null);
                EXPECT_EQUAL(system.head.next.next, null);

                /* Make sure previous pointers work correctly. */
                EXPECT_EQUAL(system.head.prev, null);
                EXPECT_EQUAL(system.head.next.prev, system.head);

                /* Make sure the particles are in the right order. */
                EXPECT_EQUAL(system.head.particle.x, 1);
                EXPECT_EQUAL(system.head.next.particle.x, 2);
            }

            @TestCase(milestone = 1, name = "Can add multiple particles.") public void test5() {
                var system = new ParticleSystem();

                /* Add some particles. */
                final int kNumParticles = 5;
                for (int i = 0; i < kNumParticles; i++) {
                    Particle particle = new Particle();
                    particle.x = i;
                    system.add(particle);
                }

                EXPECT_EQUAL(system.numParticles(), kNumParticles);

                /* Confirm they're there and in the right order. */
                int numSeen = 0;
                ParticleSystem.Cell prev = null;
                ParticleSystem.Cell curr = system.head;

                /* Walk the list checking particles. */
                while (curr != null) {
                    /* x coordinate tracks which particle this is, so this is a way of
                     * checking whether we've got the particles in the right order.
                     */
                    EXPECT_EQUAL(curr.particle.x, numSeen);

                    /* Check the list wiring - we should point back to the ParticleSystem.Cell before us,
                     * or to null if there are no ParticleSystem.Cells here.
                     */
                    EXPECT_EQUAL(curr.prev, prev);

                    /* Advance forward. */
                    prev = curr;
                    curr = curr.next;
                    numSeen++;
                }

                /* Confirm we saw the right number. */
                EXPECT_EQUAL(numSeen, kNumParticles);
                EXPECT_EQUAL(system.numParticles(), kNumParticles);
            }

            @TestCase(milestone = 2, name = "Single particle gets drawn.") public void test6() {
                var system = new ParticleSystem();

                /* This ParticleCatcher will write down all the calls to
                 * drawParticle so that we can see what's drawn.
                 */
                var catcher = new ParticleCatcher();

                /* Draw all the particles. This should do nothing. */
                system.drawParticles(catcher);

                /* Make sure nothing was drawn. */
                EXPECT_EQUAL(catcher.numDrawn(), 0);

                /* Now add a particle. */
                var particle = new Particle();
                particle.x = 13.7;
                particle.y = 4.2;
                particle.color = Color.CYAN;
                system.add(particle);

                /* Draw everything again. We should see this particle. */
                system.drawParticles(catcher);
                EXPECT_EQUAL(catcher.numDrawn(), 1);
                EXPECT_EQUAL(catcher.get(0).x,     particle.x);
                EXPECT_EQUAL(catcher.get(0).y,     particle.y);
                EXPECT_EQUAL(catcher.get(0).color, particle.color);
            }

            @TestCase(milestone = 2, name = "Multiple particles are drawn in order.") public void test7() {
                var system = new ParticleSystem();

                /* Add three particles in a specific order. */
                Particle one = new Particle(), two = new Particle(), three = new Particle();
                one.color   = Color.CYAN;
                two.color   = Color.YELLOW;
                three.color = Color.MAGENTA;

                system.add(one);
                system.add(two);
                system.add(three);

                /* Draw the particles and confirm they come back in the right order. */
                var catcher = new ParticleCatcher();
                system.drawParticles(catcher);

                EXPECT_EQUAL(catcher.numDrawn(), 3);
                EXPECT_EQUAL(catcher.get(0).color, Color.CYAN);
                EXPECT_EQUAL(catcher.get(1).color, Color.YELLOW);
                EXPECT_EQUAL(catcher.get(2).color, Color.MAGENTA);
            }
            
            @TestCase(milestone = 3, name = "Can move a single particle.") public void test8() {
                var system = new ParticleSystem();

                /* The particle we'll move around. */
                var particle = new Particle();
                particle.x = 100;
                particle.y = 100;
                particle.dx = 3;
                particle.dy = -4;
                
                /* Add a copy of this particle so we can keep the original particle parameters
                 * as a reference in the test case.
                 */
                system.add(particle.clone());
                EXPECT_EQUAL(system.numParticles(), 1);

                /* Confirm we have all the right information stored. */
                EXPECT_NOT_EQUAL(system.head, null);
                EXPECT_EQUAL(system.head.particle.x,     particle.x);
                EXPECT_EQUAL(system.head.particle.y,     particle.y);
                EXPECT_EQUAL(system.head.particle.color, particle.color);
                EXPECT_EQUAL(system.head.particle.dx,    particle.dx);
                EXPECT_EQUAL(system.head.particle.dy,    particle.dy);

                /* Move the particle. */
                system.moveParticles();

                /* The particle should be in a new spot with the same initial velocity. */
                EXPECT_EQUAL(system.head.particle.x,     particle.x + particle.dx);
                EXPECT_EQUAL(system.head.particle.y,     particle.y + particle.dy);
                EXPECT_EQUAL(system.head.particle.color, particle.color);
                EXPECT_EQUAL(system.head.particle.dx,    particle.dx);
                EXPECT_EQUAL(system.head.particle.dy,    particle.dy);

                /* Move the particle again. */
                system.moveParticles();

                /* The particle should be in a new spot with the same initial velocity. */
                EXPECT_EQUAL(system.head.particle.x,     particle.x + 2 * particle.dx);
                EXPECT_EQUAL(system.head.particle.y,     particle.y + 2 * particle.dy);
                EXPECT_EQUAL(system.head.particle.color, particle.color);
                EXPECT_EQUAL(system.head.particle.dx,    particle.dx);
                EXPECT_EQUAL(system.head.particle.dy,    particle.dy);
            }

            @TestCase(milestone = 3, name = "Can move multiple particles with different velocities.") public void test9() {
                var system = new ParticleSystem();

                /* Set up the particles. */
                Particle left = new Particle(), down = new Particle(), diag = new Particle();
                left.dx = 1;
                left.dy = 0;

                down.dx = 0;
                down.dy = 1;

                diag.dx = 1;
                diag.dy = 1;

                system.add(left);
                system.add(down);
                system.add(diag);

                /* Move the particles multiple times and confirm they're in the right
                 * places.
                 */
                final int kNumSteps = 50;
                for (int i = 0; i < kNumSteps; i++) {
                    /* Draw the particles to see where they are. */
                    var catcher = new ParticleCatcher();
                    system.drawParticles(catcher);

                    /* Make sure we see all three particles and that they're in the right
                     * places.
                     */
                    EXPECT_EQUAL(catcher.numDrawn(), 3);

                    /* Left */
                    EXPECT_EQUAL(catcher.get(0).x, i);
                    EXPECT_EQUAL(catcher.get(0).y, 0);

                    /* Down */
                    EXPECT_EQUAL(catcher.get(1).x, 0);
                    EXPECT_EQUAL(catcher.get(1).y, i);

                    /* Diag */
                    EXPECT_EQUAL(catcher.get(2).x, i);
                    EXPECT_EQUAL(catcher.get(2).y, i);

                    /* Move the particles for the next iteration. */
                    system.moveParticles();
                }
            }
            
            @TestCase(milestone = 4, name = "Don't add invalid particles to the list.") public void test10() {
                var system = new ParticleSystem();

                /* Add an unobjectionable particle. */
                Particle good = new Particle();
                good.x = 3;
                good.y = 5;
                good.color = Color.BLUE;
                good.lifetime = 137;

                /* Add the particle; confirm it's there. */
                system.add(good);
                EXPECT_NOT_EQUAL(system.head, null);
                EXPECT_EQUAL(system.head.particle.x,        good.x);
                EXPECT_EQUAL(system.head.particle.y,        good.y);
                EXPECT_EQUAL(system.head.particle.color,    good.color);
                EXPECT_EQUAL(system.head.particle.lifetime, good.lifetime);
                EXPECT_EQUAL(system.numParticles(), 1);

                /* Now make a mix of bad particles that are out of bounds. */
                Particle bad;

                bad = good.clone();
                bad.x = -0.001; // x too low
                system.add(bad);
                EXPECT_EQUAL(system.numParticles(), 1);

                bad = good.clone();
                bad.y = -0.001; // y too low
                system.add(bad);
                EXPECT_EQUAL(system.numParticles(), 1);

                bad = good.clone();
                bad.x = ParticleSystem.SCENE_WIDTH; // x too high
                system.add(bad);
                EXPECT_EQUAL(system.numParticles(), 1);

                bad = good.clone();
                bad.y = ParticleSystem.SCENE_HEIGHT; // y too high
                system.add(bad);
                EXPECT_EQUAL(system.numParticles(), 1);

                bad = good.clone();
                bad.lifetime = -1; // Negative lifetime
                system.add(bad);
                EXPECT_EQUAL(system.numParticles(), 1);

                /* Make sure the first particle is still there and unchanged. */
                EXPECT_EQUAL(system.head.particle.x,        good.x);
                EXPECT_EQUAL(system.head.particle.y,        good.y);
                EXPECT_EQUAL(system.head.particle.color,    good.color);
                EXPECT_EQUAL(system.head.particle.lifetime, good.lifetime);

                /* Make sure the list is still valid. */
                EXPECT_EQUAL(system.head.next, null);
                EXPECT_EQUAL(system.head.prev, null);
            }

            @TestCase(milestone = 4, name = "Particle removed when it leaves the screen.") public void test11() {
                var system = new ParticleSystem();

                /* A particle moving out of the world. */
                Particle escapee = new Particle();
                escapee.x = 1.5;
                escapee.y = 1.5;
                escapee.dx = -1;
                escapee.dy = -1;
                escapee.lifetime = 1000;
                system.add(escapee);

                /* Use a ParticleCatcher to see which particles are drawn. */
                var catcher = new ParticleCatcher();

                /* Initially, we should find our particle where we created it. */
                system.drawParticles(catcher);

                EXPECT_EQUAL(catcher.numDrawn(), 1);
                EXPECT_EQUAL(catcher.get(0).x, 1.5);
                EXPECT_EQUAL(catcher.get(0).y, 1.5);
                catcher.reset(); // Forget we saw the particle

                /* Moving the particle one step should leave the particle alive. */
                system.moveParticles();
                system.drawParticles(catcher);

                EXPECT_EQUAL(catcher.numDrawn(), 1);
                EXPECT_EQUAL(catcher.get(0).x, 0.5);
                EXPECT_EQUAL(catcher.get(0).y, 0.5);
                catcher.reset(); // Forget we saw the particle

                /* Moving the particle once more should remove the particle because
                 * it'll be at position (-0.5, -0.5).
                 */
                system.moveParticles();
                system.drawParticles(catcher);

                EXPECT_EQUAL(catcher.numDrawn(), 0);
                EXPECT_EQUAL(system.numParticles(), 0);
            }

            @TestCase(milestone = 4, name = "Particle removed when its lifetime ends.") public void test12() {
                var system = new ParticleSystem();

                /* A particle moving out of the world. */
                Particle timeout = new Particle();
                timeout.x = 1.5;
                timeout.y = 1.5;
                timeout.dx = 1;
                timeout.dy = 1;
                timeout.lifetime = 1;
                system.add(timeout);

                /* Use a ParticleCatcher to see which particles are drawn. */
                var catcher = new ParticleCatcher();

                /* Initially, we should find our particle where we created it. */
                system.drawParticles(catcher);

                EXPECT_EQUAL(catcher.numDrawn(), 1);
                EXPECT_EQUAL(catcher.get(0).x, 1.5);
                EXPECT_EQUAL(catcher.get(0).y, 1.5);
                catcher.reset(); // Forget we saw the particle

                /* Moving the particle one step should leave the particle alive but with
                 * a lifetime of 0.
                 */
                system.moveParticles();
                system.drawParticles(catcher);

                EXPECT_EQUAL(catcher.numDrawn(), 1);
                EXPECT_EQUAL(catcher.get(0).x, 2.5);
                EXPECT_EQUAL(catcher.get(0).y, 2.5);
                catcher.reset(); // Forget we saw the particle

                /* Moving the particle once more should remove the particle because its
                 * lifetime becomes negative.
                 */
                system.moveParticles();
                system.drawParticles(catcher);

                EXPECT_EQUAL(catcher.numDrawn(), 0);
                EXPECT_EQUAL(system.numParticles(), 0);
            }

            @TestCase(milestone = 4, name = "All particles move even if first needs to be removed.") public void test13() {
                /* Create a list of five particles at positions (0, 0), (1, 0), ... (4, 0).
                 * All particles move downward. All particles have a long lifetime except
                 * for the first, which has a lifetime of 0 and thus disappears as soon as
                 * it moves. We should see all the others shifted down by one spot.
                 */
                var system = new ParticleSystem();

                /* Used to identify particles. */
                var colors = new Color[] {
                    Color.BLACK, Color.WHITE, Color.RED, Color.GREEN, Color.YELLOW
                };

                /* Initialize the particles. */
                List<Particle> particles = new ArrayList<>();
                for (int i = 0; i < 5; i++) {
                    Particle particle = new Particle();
                    particle.x = i;
                    particle.dx = 0;
                    particle.dy = 1;
                    particle.lifetime = 100;
                    particle.color = colors[i]; // So we can flag it later
                    particles.add(particle);
                }
                particles.get(0).lifetime = 0;

                /* Add the particles. */
                for (var particle: particles) {
                    system.add(particle);
                }
                EXPECT_EQUAL(system.numParticles(), 5);

                /* Move them all. */
                system.moveParticles();
                EXPECT_EQUAL(system.numParticles(), 4); // One was removed

                /* See where they are. */
                var catcher = new ParticleCatcher();
                system.drawParticles(catcher);

                /* Confirm we have particles 1, 2, 3, and 4, in that order. */
                EXPECT_EQUAL(catcher.numDrawn(), 4);
                EXPECT_EQUAL(catcher.get(0), new DrawnParticle(1, 1, colors[1]));
                EXPECT_EQUAL(catcher.get(1), new DrawnParticle(2, 1, colors[2]));
                EXPECT_EQUAL(catcher.get(2), new DrawnParticle(3, 1, colors[3]));
                EXPECT_EQUAL(catcher.get(3), new DrawnParticle(4, 1, colors[4]));

                /* Check the linked list to make sure the wiring is right. */
                ParticleSystem.Cell curr = system.head;
                ParticleSystem.Cell prev = null;
                for (int i = 0; i < 4; i++) {
                    EXPECT_NOT_EQUAL(curr, null);
                    EXPECT_EQUAL(curr.prev, prev);

                    prev = curr;
                    curr = curr.next;
                }
                EXPECT_EQUAL(curr, null);
            }

            @TestCase(milestone = 4, name = "All particles move even if second needs to be removed.") public void test14() {
                /* Create a list of five particles at positions (0, 0), (1, 0), ... (4, 0).
                 * All particles move downward. All particles have a long lifetime except
                 * for the second, which has a lifetime of 0 and thus disappears as soon as
                 * it moves. We should see all the others shifted down by one spot.
                 */
                var system = new ParticleSystem();

                /* Used to identify particles. */
                Color[] colors = new Color[] {
                    Color.BLACK, Color.WHITE, Color.RED, Color.GREEN, Color.YELLOW
                };

                /* Initialize the particles. */
                List<Particle> particles = new ArrayList<>();
                for (int i = 0; i < 5; i++) {
                    Particle particle = new Particle();
                    particle.x = i;
                    particle.dx = 0;
                    particle.dy = 1;
                    particle.lifetime = 100;
                    particle.color = colors[i]; // So we can flag it later
                    particles.add(particle);
                }
                particles.get(1).lifetime = 0;

                /* Add the particles. */
                for (var particle: particles) {
                    system.add(particle);
                }
                EXPECT_EQUAL(system.numParticles(), 5);

                /* Move them all. */
                system.moveParticles();
                EXPECT_EQUAL(system.numParticles(), 4); // One was removed

                /* See where they are. */
                var catcher = new ParticleCatcher();
                system.drawParticles(catcher);

                /* Confirm we have particles 0, 2, 3, and 4, in that order. */
                EXPECT_EQUAL(catcher.numDrawn(), 4);
                EXPECT_EQUAL(catcher.get(0), new DrawnParticle(0, 1, colors[0]));
                EXPECT_EQUAL(catcher.get(1), new DrawnParticle(2, 1, colors[2]));
                EXPECT_EQUAL(catcher.get(2), new DrawnParticle(3, 1, colors[3]));
                EXPECT_EQUAL(catcher.get(3), new DrawnParticle(4, 1, colors[4]));

                /* Check the linked list to make sure the wiring is right. */
                ParticleSystem.Cell curr = system.head;
                ParticleSystem.Cell prev = null;
                for (int i = 0; i < 4; i++) {
                    EXPECT_NOT_EQUAL(curr, null);
                    EXPECT_EQUAL(curr.prev, prev);

                    prev = curr;
                    curr = curr.next;
                }
                EXPECT_EQUAL(curr, null);
            }

            @TestCase(milestone = 4, name = "All particles move even if last needs to be removed.") public void test15() {
                /* Create a list of five particles at positions (0, 0), (1, 0), ... (4, 0).
                 * All particles move downward. All particles have a long lifetime except
                 * for the last, which has a lifetime of 0 and thus disappears as soon as
                 * it moves. We should see all the others shifted down by one spot.
                 */
                var system = new ParticleSystem();

                /* Used to identify particles. */
                Color[] colors = new Color[] {
                    Color.BLACK, Color.WHITE, Color.RED, Color.GREEN, Color.YELLOW
                };

                /* Initialize the particles. */
                List<Particle> particles = new ArrayList<>();
                for (int i = 0; i < 5; i++) {
                    Particle particle = new Particle();
                    particle.x = i;
                    particle.dx = 0;
                    particle.dy = 1;
                    particle.lifetime = 100;
                    particle.color = colors[i]; // So we can flag it later
                    particles.add(particle);
                }
                particles.get(4).lifetime = 0;

                /* Add the particles. */
                for (var particle: particles) {
                    system.add(particle);
                }
                EXPECT_EQUAL(system.numParticles(), 5);

                /* Move them all. */
                system.moveParticles();
                EXPECT_EQUAL(system.numParticles(), 4); // One was removed

                /* See where they are. */
                var catcher = new ParticleCatcher();
                system.drawParticles(catcher);

                /* Confirm we have particles 0, 1, 2, and 3, in that order. */
                EXPECT_EQUAL(catcher.numDrawn(), 4);
                EXPECT_EQUAL(catcher.get(0), new DrawnParticle(0, 1, colors[0]));
                EXPECT_EQUAL(catcher.get(1), new DrawnParticle(1, 1, colors[1]));
                EXPECT_EQUAL(catcher.get(2), new DrawnParticle(2, 1, colors[2]));
                EXPECT_EQUAL(catcher.get(3), new DrawnParticle(3, 1, colors[3]));

                /* Check the linked list to make sure the wiring is right. */
                ParticleSystem.Cell curr = system.head;
                ParticleSystem.Cell prev = null;
                for (int i = 0; i < 4; i++) {
                    EXPECT_NOT_EQUAL(curr, null);
                    EXPECT_EQUAL(curr.prev, prev);

                    prev = curr;
                    curr = curr.next;
                }
                EXPECT_EQUAL(curr, null);
            }

            @TestCase(milestone = 4, name = "All particles move even if second-to-last needs to be removed.") public void test16() {
                /* Create a list of five particles at positions (0, 0), (1, 0), ... (4, 0).
                 * All particles move downward. All particles have a long lifetime except
                 * for the penultimate, which has a lifetime of 0 and thus disappears as soon as
                 * it moves. We should see all the others shifted down by one spot.
                 */
                var system = new ParticleSystem();

                /* Used to identify particles. */
                Color[] colors = new Color[] {
                    Color.BLACK, Color.WHITE, Color.RED, Color.GREEN, Color.YELLOW
                };

                /* Initialize the particles. */
                List<Particle> particles = new ArrayList<>();
                for (int i = 0; i < 5; i++) {
                    Particle particle = new Particle();
                    particle.x = i;
                    particle.dx = 0;
                    particle.dy = 1;
                    particle.lifetime = 100;
                    particle.color = colors[i]; // So we can flag it later
                    particles.add(particle);
                }
                particles.get(3).lifetime = 0;

                /* Add the particles. */
                for (var particle: particles) {
                    system.add(particle);
                }
                EXPECT_EQUAL(system.numParticles(), 5);

                /* Move them all. */
                system.moveParticles();
                EXPECT_EQUAL(system.numParticles(), 4); // One was removed

                /* See where they are. */
                var catcher = new ParticleCatcher();
                system.drawParticles(catcher);

                /* Confirm we have particles 0, 1, 2, and 4, in that order. */
                EXPECT_EQUAL(catcher.numDrawn(), 4);
                EXPECT_EQUAL(catcher.get(0), new DrawnParticle(0, 1, colors[0]));
                EXPECT_EQUAL(catcher.get(1), new DrawnParticle(1, 1, colors[1]));
                EXPECT_EQUAL(catcher.get(2), new DrawnParticle(2, 1, colors[2]));
                EXPECT_EQUAL(catcher.get(3), new DrawnParticle(4, 1, colors[4]));

                /* Check the linked list to make sure the wiring is right. */
                ParticleSystem.Cell curr = system.head;
                ParticleSystem.Cell prev = null;
                for (int i = 0; i < 4; i++) {
                    EXPECT_NOT_EQUAL(curr, null);
                    EXPECT_EQUAL(curr.prev, prev);

                    prev = curr;
                    curr = curr.next;
                }
                EXPECT_EQUAL(curr, null);
            }

            @TestCase(milestone = 4, name = "All particles move even if first needs to be removed.") public void test17() {
                /* Create a list of five particles at positions (0, 0), (1, 0), ... (4, 0).
                 * All particles move downward. All particles have a long lifetime except
                 * for the first, which has a lifetime of 0 and thus disappears as soon as
                 * it moves. We should see all the others shifted down by one spot.
                 */
                var system = new ParticleSystem();

                /* Used to identify particles. */
                var colors = new Color[] {
                    Color.BLACK, Color.WHITE, Color.RED, Color.GREEN, Color.YELLOW
                };

                /* Initialize the particles. */
                List<Particle> particles = new ArrayList<>();
                for (int i = 0; i < 5; i++) {
                    Particle particle = new Particle();
                    particle.x = i;
                    particle.dx = 0;
                    particle.dy = 1;
                    particle.lifetime = 100;
                    particle.color = colors[i]; // So we can flag it later
                    particles.add(particle);
                }
                particles.get(2).lifetime = 0;

                /* Add the particles. */
                for (var particle: particles) {
                    system.add(particle);
                }
                EXPECT_EQUAL(system.numParticles(), 5);

                /* Move them all. */
                system.moveParticles();
                EXPECT_EQUAL(system.numParticles(), 4); // One was removed

                /* See where they are. */
                var catcher = new ParticleCatcher();
                system.drawParticles(catcher);

                /* Confirm we have particles 0, 1, 3, and 4, in that order. */
                EXPECT_EQUAL(catcher.numDrawn(), 4);
                EXPECT_EQUAL(catcher.get(0), new DrawnParticle(0, 1, colors[0]));
                EXPECT_EQUAL(catcher.get(1), new DrawnParticle(1, 1, colors[1]));
                EXPECT_EQUAL(catcher.get(2), new DrawnParticle(3, 1, colors[3]));
                EXPECT_EQUAL(catcher.get(3), new DrawnParticle(4, 1, colors[4]));

                /* Check the linked list to make sure the wiring is right. */
                ParticleSystem.Cell curr = system.head;
                ParticleSystem.Cell prev = null;
                for (int i = 0; i < 4; i++) {
                    EXPECT_NOT_EQUAL(curr, null);
                    EXPECT_EQUAL(curr.prev, prev);

                    prev = curr;
                    curr = curr.next;
                }
                EXPECT_EQUAL(curr, null);
            }

            @TestCase(milestone = 4, name = "All particles move even if many need to be removed.") public void test18() {
                /* Create a list of five particles at positions (0, 0), (1, 0), ... (4, 0).
                 * All particles move downward. Particles 0, 2, and 4 have lifetime 0 and
                 * thus will disappear after the first step. The other two will live for
                 * multiple time steps.
                 */
                var system = new ParticleSystem();

                /* Used to identify particles. */
                var colors = new Color[] {
                    Color.BLACK, Color.WHITE, Color.RED, Color.GREEN, Color.YELLOW
                };

                /* Initialize the particles. */
                List<Particle> particles = new ArrayList<>();
                for (int i = 0; i < 5; i++) {
                    Particle particle = new Particle();
                    particle.x = i;
                    particle.dx = 0;
                    particle.dy = 1;
                    particle.lifetime = 100;
                    particle.color = colors[i]; // So we can flag it later
                    particles.add(particle);
                }
                particles.get(0).lifetime = 0;
                particles.get(2).lifetime = 0;
                particles.get(4).lifetime = 0;

                /* Add the particles. */
                for (Particle particle: particles) {
                    system.add(particle);
                }
                EXPECT_EQUAL(system.numParticles(), 5);

                /* Move them all. */
                system.moveParticles();
                EXPECT_EQUAL(system.numParticles(), 2); // Three were removed

                /* See where they are. */
                var catcher = new ParticleCatcher();
                system.drawParticles(catcher);

                /* Confirm we have particles 1 and 3, in that order. */
                EXPECT_EQUAL(catcher.numDrawn(), 2);
                EXPECT_EQUAL(catcher.get(0), new DrawnParticle(1, 1, colors[1]));
                EXPECT_EQUAL(catcher.get(1), new DrawnParticle(3, 1, colors[3]));

                /* Check the linked list to make sure the wiring is right. */
                ParticleSystem.Cell curr = system.head;
                ParticleSystem.Cell prev = null;
                for (int i = 0; i < 2; i++) {
                    EXPECT_NOT_EQUAL(curr, null);
                    EXPECT_EQUAL(curr.prev, prev);

                    prev = curr;
                    curr = curr.next;
                }
                EXPECT_EQUAL(curr, null);
            }

            @TestCase(milestone = 4, name = "After all particles expire, can add new particles.") public void test19() {
                var system = new ParticleSystem();

                /* On entry to this loop, the particle system should be empty.
                 * On exit, it should be empty.
                 */
                for (int round = 0; round < 10; round++) {
                    /* Shouldn't see anything. */
                    var catcher = new ParticleCatcher();
                    system.drawParticles(catcher);
                    EXPECT_EQUAL(catcher.numDrawn(), 0);

                    /* Create a bunch of short-lived particles. */
                    for (int i = 0; i < 5; i++) {
                        Particle p = new Particle();
                        p.x = i;
                        p.y = i;
                        p.lifetime = 2;
                        system.add(p);
                    }
                    EXPECT_EQUAL(system.numParticles(), 5);

                    /* Should see five particles. */
                    system.drawParticles(catcher);
                    EXPECT_EQUAL(catcher.numDrawn(), 5);

                    /* They should be in the right place. */
                    for (int i = 0; i < 5; i++) {
                        EXPECT_EQUAL(catcher.get(i).x, i);
                        EXPECT_EQUAL(catcher.get(i).y, i);
                    }

                    /* Move the particles a bunch to clear everything out as their
                     * lifetimes expire.
                     */
                    for (int i = 0; i < 10; i++) {
                        system.moveParticles();
                    }

                    /* We should have no particles. */
                    EXPECT_EQUAL(system.numParticles(), 0);
                    catcher.reset();
                    system.drawParticles(catcher);
                    EXPECT_EQUAL(catcher.numDrawn(), 0);
                }
            }
            
            @TestCase(milestone = 5, name = "Ballistic particles accelerate downward.") public void test20() {
                var system = new ParticleSystem();

                var ballistic = new Particle();
                ballistic.type = ParticleType.BALLISTIC;
                ballistic.x = 0;
                ballistic.y = 100;

                /* Fire upward and to the right. */
                ballistic.dx = 1;
                ballistic.dy = -10;

                system.add(ballistic);

                /* Expected positions. */
                var positions = new Point2D.Double[] {
                    new Point2D.Double(  0, 100 ),
                    new Point2D.Double(  1,  90 ), // y += -10
                    new Point2D.Double(  2,  81 ), // y += -9
                    new Point2D.Double(  3,  73 ), // y += -8
                    new Point2D.Double(  4,  66 ),
                    new Point2D.Double(  5,  60 ),
                    new Point2D.Double(  6,  55 ),
                    new Point2D.Double(  7,  51 ),
                    new Point2D.Double(  8,  48 ),
                    new Point2D.Double(  9,  46 ),
                    new Point2D.Double( 10,  45 ),
                    new Point2D.Double( 11,  45 ), // y += 0
                    new Point2D.Double( 12,  46 ),
                    new Point2D.Double( 13,  48 ),
                    new Point2D.Double( 14,  51 ),
                    new Point2D.Double( 15,  55 ),
                    new Point2D.Double( 16,  60 ),
                    new Point2D.Double( 17,  66 ),
                    new Point2D.Double( 18,  73 ),
                    new Point2D.Double( 19,  81 ), // y += 8
                    new Point2D.Double( 20,  90 ), // y += 9
                    new Point2D.Double( 21, 100 ), // y += 10
                };

                /* Repeatedly check positions and move forward. */
                for (int i = 0; i < positions.length; i++) {
                    var catcher = new ParticleCatcher();
                    system.drawParticles(catcher);
                    EXPECT_EQUAL(catcher.numDrawn(), 1);
                    EXPECT_EQUAL(catcher.get(0).x, positions[i].x);
                    EXPECT_EQUAL(catcher.get(0).y, positions[i].y);

                    system.moveParticles();
                }
            }

            @TestCase(milestone = 5, name = "Fireworks accelerate downward.") public void test21() {
                var system = new ParticleSystem();

                var firework = new Particle();
                firework.type = ParticleType.FIREWORK;
                firework.x = 0;
                firework.y = 100;

                /* Fire upward and to the right. */
                firework.dx = 1;
                firework.dy = -10;

                system.add(firework);

                /* Expected positions. */
                var positions = new Point2D.Double[] {
                    new Point2D.Double(  0, 100 ),
                    new Point2D.Double(  1,  90 ), // y += -10
                    new Point2D.Double(  2,  81 ), // y += -9
                    new Point2D.Double(  3,  73 ), // y += -8
                    new Point2D.Double(  4,  66 ),
                    new Point2D.Double(  5,  60 ),
                    new Point2D.Double(  6,  55 ),
                    new Point2D.Double(  7,  51 ),
                    new Point2D.Double(  8,  48 ),
                    new Point2D.Double(  9,  46 ),
                    new Point2D.Double( 10,  45 ),
                    new Point2D.Double( 11,  45 ), // y += 0
                    new Point2D.Double( 12,  46 ),
                    new Point2D.Double( 13,  48 ),
                    new Point2D.Double( 14,  51 ),
                    new Point2D.Double( 15,  55 ),
                    new Point2D.Double( 16,  60 ),
                    new Point2D.Double( 17,  66 ),
                    new Point2D.Double( 18,  73 ),
                    new Point2D.Double( 19,  81 ), // y += 8
                    new Point2D.Double( 20,  90 ), // y += 9
                    new Point2D.Double( 21, 100 ), // y += 10
                };

                /* Repeatedly check positions and move forward. */
                for (int i = 0; i < positions.length; i++) {
                    var catcher = new ParticleCatcher();
                    system.drawParticles(catcher);
                    EXPECT_EQUAL(catcher.numDrawn(), 1);
                    EXPECT_EQUAL(catcher.get(0).x, positions[i].x);
                    EXPECT_EQUAL(catcher.get(0).y, positions[i].y);

                    system.moveParticles();
                }
            }

            @TestCase(milestone = 5, name = "Fireworks explode when timer expires.") public void test22() {
                var system = new ParticleSystem();

                /* Create and add a firework. */
                var firework = new Particle();
                firework.type = ParticleType.FIREWORK;
                firework.lifetime = 0; // Explodes as soon as it moves
                firework.x = 100;
                firework.y = 100;
                system.add(firework);

                EXPECT_EQUAL(system.numParticles(), 1);

                /* Move the particles. This should trigger the explosion. */
                system.moveParticles();

                /* There should be 50 new particles, all of which are streamers, and
                 * all of which are the same color.
                 */
                EXPECT_EQUAL(system.numParticles(), 50);
                EXPECT_NOT_EQUAL(system.head, null);

                /* Color of the first particle. */
                Color color = system.head.particle.color;
                for (var curr = system.head; curr != null; curr = curr.next) {
                    EXPECT_EQUAL(curr.particle.color, color);
                }
            }

        }.runTests();   
    }
}
