package particles.gui;
import particles.*;
import particles.scenes.*;
import javax.swing.*;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.Polygon;
import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.event.*;
import java.util.List;

/* Component that renders a Scene. Handles all the logic to recenter everything within the screen,
 * draw the border, interact with the mouse, etc.
 */
public class ParticleDisplay extends JPanel {
    private Scene scene;
    private final Timer timer;
    
    /* Mouse listener type that converts from screen coordinates to logical coordinates
     * so that Scenes don't have to do all those conversions.
     */
    private class MouseHandler implements MouseListener, MouseMotionListener {
        private Point2D.Double toWorld(Point2D.Double pt) {
            return new Point2D.Double(pt.x - (getWidth()  - ParticleSystem.SCENE_WIDTH)  / 2.0,
                                      pt.y - (getHeight() - ParticleSystem.SCENE_HEIGHT) / 2.0);
        }
        
        @Override public void mouseClicked(MouseEvent e) {
            // Not forwarded
        }
        
        @Override public void mouseEntered(MouseEvent e) {
            // Not forwarded
        }
        
        @Override public void mouseExited(MouseEvent e) {
            // Not forwarded
        }
        
        @Override public void mousePressed(MouseEvent e) {
            if (scene == null) return;
            
            Point2D.Double pt = toWorld(new Point2D.Double(e.getX(), e.getY()));
            scene.mousePressed(pt.x, pt.y);
        }
        
        @Override public void mouseReleased(MouseEvent e) {
            if (scene == null) return;
                    
            Point2D.Double pt = toWorld(new Point2D.Double(e.getX(), e.getY()));
            scene.mouseReleased(pt.x, pt.y);
        }
        
        @Override public void mouseMoved(MouseEvent e) {
            if (scene == null) return;
        
            Point2D.Double pt = toWorld(new Point2D.Double(e.getX(), e.getY()));
            scene.mouseMoved(pt.x, pt.y);
        }
        
        @Override public void mouseDragged(MouseEvent e) {
            if (scene == null) return;
        
            Point2D.Double pt = toWorld(new Point2D.Double(e.getX(), e.getY()));
            scene.mouseDragged(pt.x, pt.y);
        }
    }
    
    public ParticleDisplay() {
        /* Big enough for the scene plus a border. */
        setPreferredSize(new Dimension((int) ParticleSystem.SCENE_WIDTH + 100, (int) ParticleSystem.SCENE_HEIGHT + 100));
        
        /* Timer that ticks and updates scenes. */
        timer = new Timer(50, null);
        timer.addActionListener((ActionEvent e) -> {
            if (scene != null) {
                /* Update the frame rate in case the scene changed it. */
                timer.setDelay(1000 / scene.getFramesPerSecond());
                scene.tick();
                repaint();
            }
        });
        
        timer.start();
        
        /* React to the mouse. */
        MouseHandler handler = new MouseHandler();
        addMouseListener(handler);
        addMouseMotionListener(handler);
    }
    
    /* Sets which scene is being displayed. */
    public void setScene(Scene s) {
        scene = s;
        repaint();
    }
    
    @Override public void paint(Graphics g) {    
        /* Wipe everything that's already there. */
        g.setColor(Color.WHITE);
        g.fillRect(0, 0, getWidth(), getHeight());

        /* Render the scene. */
        if (scene != null) {
            scene.draw(new ActualGraphics(g));
        }

        /* Draw four rectangles around the content area to ensure that nothing
         * overdraws out of the allotted area.
         */
        g.setColor(Color.WHITE);

        int xMargin = (int) ((getWidth()  - ParticleSystem.SCENE_WIDTH)  / 2);
        int yMargin = (int) ((getHeight() - ParticleSystem.SCENE_HEIGHT) / 2);

        /* Top */
        g.fillRect(0, 0, getWidth(), yMargin);

        /* Left */
        g.fillRect(0, 0, xMargin, getHeight());

        /* Right */
        g.fillRect(getWidth()  - xMargin, 0, xMargin, getHeight());

        /* Bottom. */
        g.fillRect(0, getHeight() - yMargin, getWidth(),  yMargin);

        /* Draw a box around the scene. */
        g.setColor(Color.BLACK);
        g.drawRect((int) ((getWidth()  - ParticleSystem.SCENE_WIDTH)  / 2.0),
                   (int) ((getHeight() - ParticleSystem.SCENE_HEIGHT) / 2.0),
                   (int) ParticleSystem.SCENE_WIDTH, (int) ParticleSystem.SCENE_HEIGHT);
                   
        /* Sync with the OS so that animations look smooth. */
        Toolkit.getDefaultToolkit().sync();
    }
    
    /* The actual graphics system used by ParticleGraphics. This converts everything from
     * scene coordinates to screen coordinates before drawing so that everything is
     * always centered.
     */
    private class ActualGraphics extends ParticleGraphics {
        private Graphics graphics;
        
        private static final double PARTICLE_RADIUS = 2;
        
        public ActualGraphics(Graphics g) {
            graphics = g;
        }
    
        /* Local -> Global conversion. */
        private Point2D.Double toScreen(Point2D.Double pt) {
            return new Point2D.Double(pt.x + (getWidth()  - ParticleSystem.SCENE_WIDTH)  / 2.0,
                                      pt.y + (getHeight() - ParticleSystem.SCENE_HEIGHT) / 2.0);
        }
        
        private Rectangle2D.Double toScreen(Rectangle2D.Double r) {
            return new Rectangle2D.Double(r.x + (getWidth()  - ParticleSystem.SCENE_WIDTH)  / 2.0,
                                          r.y + (getHeight() - ParticleSystem.SCENE_HEIGHT) / 2.0,
                                          r.width, r.height);
        }
        
        @Override public void setColor(Color color) {
            graphics.setColor(color);
        }
        @Override public Color getColor() {
            return graphics.getColor();
        }
        
        @Override public void drawParticle(double x, double y) {        
            fillOval(x - PARTICLE_RADIUS, y - PARTICLE_RADIUS, 2 * PARTICLE_RADIUS, 2 * PARTICLE_RADIUS);
        }
        
        @Override public void drawRect(double x, double y, double width, double height) {
            Rectangle2D.Double bounds = toScreen(new Rectangle2D.Double(x, y, width, height));
            graphics.drawRect((int) bounds.x, (int) bounds.y, (int) bounds.width, (int) bounds.height);
        }
        
        @Override public void fillRect(double x, double y, double width, double height) {
            Rectangle2D.Double bounds = toScreen(new Rectangle2D.Double(x, y, width, height));
            graphics.fillRect((int) bounds.x, (int) bounds.y, (int) bounds.width, (int) bounds.height);
        }
        
        @Override public void drawOval(double x, double y, double width, double height) {
            Rectangle2D.Double bounds = toScreen(new Rectangle2D.Double(x, y, width, height));
            graphics.drawOval((int) bounds.x, (int) bounds.y, (int) bounds.width, (int) bounds.height);
        }
        
        @Override public void fillOval(double x, double y, double width, double height) {        
            Rectangle2D.Double bounds = toScreen(new Rectangle2D.Double(x, y, width, height));            
            graphics.fillOval((int) bounds.x, (int) bounds.y, (int) bounds.width, (int) bounds.height);
        }
        
        private Polygon toPolygon(List<Point2D.Double> pts) {
            Polygon result = new Polygon();
            
            /* Translate to screen coordinates. */
            for (Point2D.Double pt: pts) {
                Point2D.Double coord = toScreen(pt);
                result.addPoint((int) coord.x, (int) coord.y);
            }
            
            return result;
        }
        
        @Override public void drawPoly(List<Point2D.Double> pts) {
            graphics.drawPolygon(toPolygon(pts));
        }
        
        @Override public void fillPoly(List<Point2D.Double> pts) {
            graphics.fillPolygon(toPolygon(pts));
        }
        
        @Override public void drawLine(double x0, double y0, double x1, double y1) {
            Point2D.Double from = toScreen(new Point2D.Double(x0, y0));
            Point2D.Double to   = toScreen(new Point2D.Double(x1, y1));
            graphics.drawLine((int) from.x, (int) from.y, (int) to.x, (int) to.y);
        }
    }
}
