package particles.gui;
import particles.*;
import particles.scenes.*;
import javax.swing.*;
import java.io.*;
import java.util.*;
import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.event.*;
import java.lang.reflect.*;

/* Program that lets the user select scenes and watch the result. */
public class ParticleGUI {
    private JFrame window;
    private ParticleDisplay display;
    private JComboBox<SceneFactory> sceneSelector;
    
    /* Type that instantiates new Scenes given a Class object. */
    private static final class SceneFactory {
        private Constructor<? extends Scene> ctor;
        
        /* Either a constructor, or null for a null scene. */
        public SceneFactory(Constructor<? extends Scene> ctor) {
            this.ctor = ctor;
        }
        
        /* Factory method that makes a scene. (Not in the sense of causing a disruption;
         * in the sense of instantiating a new Scene subclass. ^_^)
         */
        public Scene makeScene() {
            /* If there's no scene, don't make a scene. */
            if (ctor == null) return null;
            
            try {
                return ctor.newInstance();
            } catch (InstantiationException e) {
                throw new RuntimeException("Error instantiating " + ctor.getDeclaringClass());
            } catch (IllegalAccessException e) {
                throw new RuntimeException("Error instantiating " + ctor.getDeclaringClass());
            } catch (IllegalArgumentException e) {
                throw new RuntimeException("Error instantiating " + ctor.getDeclaringClass());
            }  catch (InvocationTargetException e) {
                throw new RuntimeException(e);
            }
        }
        
        @Override public String toString() {
            if (ctor == null) return "-";
            
            /* Base name will be something of the form 'class X.X.X.actualName. */
            var baseName = ctor.getDeclaringClass().toString();
            
            /* Skip over 'class ' and 'X.X.X.' */
            return baseName.substring(6 + SCENE_PATH.length());
        }
    }
    
    /* Path to the directory containing scene files. For Silly Internal Reasons,
     * this must not start with ./ and must end with a / character.
     */
    private static final String SCENE_PATH = "particles/scenes/";
    
    public ParticleGUI() {
        /* Main window. */
        window = new JFrame();
        window.setLayout(new BorderLayout());
        window.setTitle("Particle Systems");
        window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

        /* Main display. */   
        display = new ParticleDisplay();     
        window.add(display, BorderLayout.CENTER);

        /* Control panel. */
        window.add(makeControlPanel(), BorderLayout.SOUTH);
        
        /* Show the world our result. */
        window.pack();
        window.setVisible(true);
    }
    
    /* Builds the control panel. */
    private JPanel makeControlPanel() {
        var panel = new JPanel();
        panel.setLayout(new FlowLayout());
            
        sceneSelector = makeSceneSelector();
        panel.add(sceneSelector);
            
        var goButton = makeGoButton();
        panel.add(goButton);
        
        return panel;
    }
    
    /* Makes the drop-down selector for Scene subclasses. */
    private JComboBox<SceneFactory> makeSceneSelector() {
        /* Sort scenes alphabetically, to be nice. */
        var scenes = sceneClassesIn(SCENE_PATH);
        Collections.sort(scenes, (Constructor<? extends Scene> lhs, Constructor<? extends Scene> rhs) -> {
            var one = lhs.getDeclaringClass().toString();
            var two = rhs.getDeclaringClass().toString();
            return one.compareTo(two);
        });
    
        var result = new JComboBox<SceneFactory>();
        
        /* Add a default "no scene" option. */
        result.addItem(new SceneFactory(null));
        
        /* Now add the actual scenes. */
        for (var clazz: scenes) {
            result.addItem(new SceneFactory(clazz));
        }
        return result;
    }
    
    /* Makes the button that loads a scene. */
    private JButton makeGoButton() {
        var result = new JButton("Load Scene");
        result.addActionListener((ActionEvent e) -> {
            display.setScene(((SceneFactory) sceneSelector.getSelectedItem()).makeScene());
        });
        return result;
    }
    
    /* Returns a list of all subclasses of Scene found in the relevant directory. */
    private List<Constructor<? extends Scene>> sceneClassesIn(String path) {
        List<Constructor<? extends Scene>> result = new ArrayList<>();
        for (File file: classFilesIn(path)) {
            /* Trim off ".class" */
            String className = file.getName().substring(0, file.getName().length() - 6);
                
            /* Prepend package, if appropriate. */
            if (!path.equals("")) {
                className = path.replace("/", ".") + className;
            }

            /* Try loading the class and see if it's a Scene. */
            try {    
                Class<?> clazz = Class.forName(className);                
                
                /* Make sure it is a Scene and has a zero-arg constructor. */
                if (Scene.class.equals(clazz.getSuperclass())) {
                    var ctor = clazz.asSubclass(Scene.class).getDeclaredConstructor();
                    result.add(ctor);
                }
                
            } catch (LinkageError e) {
                System.err.println("Warning: Linkage failed for class " + file.getName());
                System.err.println(e);
                e.printStackTrace();                
            } catch (ClassNotFoundException e) {
                System.err.println("Warning: Class " + file.getName() + " was not found.");
            } catch (ClassCastException e) {
                System.err.println("Warning: Couldn't cast " + file.getName() + " to Scene, but it has Scene as a superclass?");
            } catch (NoSuchMethodException e) {
                System.err.println("Warning: Class " + file.getName() + " is a Scene, but has no zero-arg constructor.");
            } catch (SecurityException e) {
                System.err.println("Warning: Class " + file.getName() + " is a Scene, but a security blocked its constructor.");
            }
        }
        
        return result;
    }
    
    /* Returns a list of all class files in the given directory. */
    private List<File> classFilesIn(String path) {
        File sceneDir = new File("./" + path);
        
        /* Must make sure we can find the appropriate path. */
        if (!sceneDir.exists() || !sceneDir.isDirectory()) {
            throw new RuntimeException("Cannot load scene directory.");
        }
            
        /* Find all class files in that directory. */
        List<File> result = new ArrayList<>();
        for (File file: sceneDir.listFiles()) {
            if (file.isFile() && file.getName().endsWith(".class")) {
                result.add(file);
            }
        }
        return result;
    }

    public static void main(String[] args) {
        try {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        } catch (Exception e) {
            System.err.println("Cannot set look and feel; falling back on default.");
        }
        new ParticleGUI();
    }
}
