package particles.scenes;
import particles.*;
import java.awt.Color;
import java.awt.geom.*;
import java.util.*;

public class Fountain extends Scene {
    private ParticleSystem system = new ParticleSystem();
    private List<Point2D.Double> emitters = new ArrayList<>();
    
    /* Spacing between elements of the fountain. */
    private static final double kSpacing  = SCENE_HEIGHT / 9;

    /* Thickness of the central column. */
    private static final double kColumnThickness = 20;

    /* Thickness of each branch. */
    private static final double kBranchThickness = 10;

    /* Angles at which water can be sprayed from the emitters. */
    private static final double kMinAngle = Math.PI / 2 - Math.PI / 12;
    private static final double kMaxAngle = Math.PI / 2 + Math.PI / 12;

    /* Water speed. */
    private static final double kMinWaterSpeed = 10;
    private static final double kMaxWaterSpeed = 20;

    /* Particles per emitter per frame. */
    private static final int kFlowRate = 10;

    /* Minimum and maximum white components for water particles. */
    private static final int kMinWhite = 60;
    private static final int kMaxWhite = 210;

    public Fountain() {
        /* Rough sketch of the fountain:
         *
         *            *
         *            |
         *            |
         *         *--+--*
         *            |
         *            |
         *      *--*--+--*--*
         *            |
         *            |
         *
         * Each * here is an emitter. The | and -- lines represent the same
         * spacing.
         */
        double bottomRowY = SCENE_HEIGHT - 1 - 2 * kSpacing;
        emitters.add(new Point2D.Double(SCENE_WIDTH / 2 - 2 * kSpacing, bottomRowY));
        emitters.add(new Point2D.Double(SCENE_WIDTH / 2 - 1 * kSpacing, bottomRowY));
        emitters.add(new Point2D.Double(SCENE_WIDTH / 2 + 1 * kSpacing, bottomRowY));
        emitters.add(new Point2D.Double(SCENE_WIDTH / 2 + 2 * kSpacing, bottomRowY));

        double middleRowY = SCENE_HEIGHT - 1 - 4 * kSpacing;
        emitters.add(new Point2D.Double(SCENE_WIDTH / 2 - 1 * kSpacing, middleRowY));
        emitters.add(new Point2D.Double(SCENE_WIDTH / 2 + 1 * kSpacing, middleRowY));

        double topRowY = SCENE_HEIGHT - 1 - 6 * kSpacing;
        emitters.add(new Point2D.Double(SCENE_WIDTH / 2, topRowY));
        emitters.add(new Point2D.Double(SCENE_WIDTH / 2, topRowY));
    }

    @Override public void tick() {
        /* Each source emits water. */
        for (var source: emitters) {
            /* Each source emits multiple particles. */
            for (int i = 0; i < kFlowRate; i++) {
                /* Pick an angle and fire water at that angle. */
                double theta = RandomGenerator.randomReal(kMinAngle, kMaxAngle);

                /* Select a speed. */
                double speed = RandomGenerator.randomReal(kMinWaterSpeed, kMaxWaterSpeed);

                /* dy is negative because positive y corresponds to moving down. */
                double dx =  speed * Math.cos(theta);
                double dy = -speed * Math.sin(theta);

                /* Set this up as a particle. */
                Particle data = new Particle();
                data.color = waterColor();
                data.x = source.x;
                data.y = source.y;
                data.dx = dx;
                data.dy = dy;
                data.type = ParticleType.BALLISTIC;
                data.lifetime = Integer.MAX_VALUE; // Particle only disappears when falling off the frame.

                system.add(data);
            }
        }

        system.moveParticles();
    }

    /* Picks a color for a particle of water. The particle always has the maximum
     * possible blue component and has a red/green component that is randomly
     * chosen.
     */
    private Color waterColor() {
        int white = RandomGenerator.randomInteger(kMinWhite, kMaxWhite);
        return new Color(white, white, 255);
    }

    /* Draw the particles and the fountain itself. */
    @Override public void draw(ParticleGraphics g) {
        /* First, the fountain. */
        g.setColor(Color.GRAY);

        /* Central column. */
        g.fillRect((SCENE_WIDTH - kColumnThickness) / 2,
                   SCENE_HEIGHT - 1 - 6 * kSpacing,
                   kColumnThickness,
                   6 * kSpacing);

        /* Middle branch. */
        g.fillRect((SCENE_WIDTH - 2 * kSpacing) / 2,
                   SCENE_HEIGHT - 1 - 4 * kSpacing - kBranchThickness / 2,
                   2 * kSpacing,
                   kBranchThickness);

        /* Bottom branch. */
        g.fillRect((SCENE_WIDTH - 4 * kSpacing) / 2,
                   SCENE_HEIGHT - 1 - 2 * kSpacing - kBranchThickness / 2,
                   4 * kSpacing,
                   kBranchThickness);

        system.drawParticles(g);
    }
}
