package particles.scenes;
import particles.*;
import java.awt.Color;
import java.awt.geom.*;

public class MagicWand extends Scene {
    private ParticleSystem system = new ParticleSystem();
    
    /* Is the mouse down and, if so, where? */
    private boolean mouseDown = false;
    private Point2D.Double mouse = new Point2D.Double(SCENE_WIDTH / 2.0, SCENE_HEIGHT / 2.0);
    
    /* Size of the tip of the magic wand. */
    private static final double kTipWidth = 10;
    private static final double kTipHeight = 20;

    /* How tall the wand body - excluding the tip - is. */
    private static final double kWandHeight = 100;

    /* How many particles to create per frame when the mouse is down. */
    private static final int kDownRate = 50;

    /* How many particles to create when the mouse moves. */
    private static final int kMoveRate = 1;

    /* Speed of a particle. */
    private static final double kMinStreamerSpeed = 3;
    private static final double kMaxStreamerSpeed = 20;

    /* Lifetime of a particle. */
    private static final int kMinLifetime = 0;
    private static final int kMaxLifetime = 20;

    /* Speed controls for a falling particle. */
    private static final double kMinMoveX = -3;
    private static final double kMaxMoveX = +3;

    private static final double kMinMoveY = -6;
    private static final double kMaxMoveY = -3;
    
    @Override public void tick() {
        /* If the mouse is down, create a shower of particles from the mouse
         * position - which is also where the tip of the magic wand is.
         */
        if (mouseDown) {
            for (int i = 0; i < kDownRate; i++) {
                /* Random angle / speed to fire the particle. */
                double theta = RandomGenerator.randomReal(0, 2 * Math.PI);
                double speed = RandomGenerator.randomReal(kMinStreamerSpeed, kMaxStreamerSpeed);

                /* How long the particle lives for. */
                int lifetime  = RandomGenerator.randomInteger(kMinLifetime, kMaxLifetime);

                Particle particle = new Particle();

                /* Center on the mouse. */
                particle.x = mouse.x;
                particle.y = mouse.y;

                /* Polar coordinates to the rescue! */
                particle.dx = speed * Math.cos(theta);
                particle.dy = speed * Math.sin(theta);

                particle.lifetime = lifetime;
                particle.type = ParticleType.STREAMER;
                particle.color = RandomGenerator.randomColor();

                system.add(particle);
            }
        }

        system.moveParticles();
    }
    
    @Override public void draw(ParticleGraphics graphics) {
        system.drawParticles(graphics);

        /* Draw the tip of the wand. */
        graphics.setColor(Color.WHITE);
        graphics.fillRect(mouse.x - kTipWidth / 2, mouse.y - kTipHeight / 2, kTipWidth, kTipHeight);

        /* Draw the body of the wand, and the wand border. */
        graphics.setColor(Color.BLACK);
        graphics.drawRect(mouse.x - kTipWidth / 2, mouse.y - kTipHeight / 2, kTipWidth, kTipHeight);
        graphics.fillRect(mouse.x - kTipWidth / 2, mouse.y + kTipHeight / 2, kTipWidth, kWandHeight);
    }
    
    /* Mouse motion. In respond to mouse events, remember where the mouse is. */
    @Override public void mouseMoved(double x, double y) {
        /* Remember where the mouse is. */
        mouse = new Point2D.Double(x, y);

        /* Create particles that emanates from the mouse position. */
        for (int i = 0; i < kMoveRate; i++) {
            Particle particle = new Particle();
            particle.x = x;
            particle.y = y;

            particle.dx = RandomGenerator.randomReal(kMinMoveX, kMaxMoveX);
            particle.dy = RandomGenerator.randomReal(kMinMoveY, kMaxMoveY);

            particle.lifetime = Integer.MAX_VALUE; // Live forever, basically
            particle.color = RandomGenerator.randomColor();

            particle.type = ParticleType.BALLISTIC;
            system.add(particle);
        }
    }

    @Override public void mousePressed(double x, double y) {
        /* Record that the mouse is down. */
        mouseDown = true;
        mouse = new Point2D.Double(x, y);
    }

    @Override public void mouseReleased(double x, double y) {
        /* Record that the mouse is up. */
        mouseDown = false;
    }

    @Override public void mouseDragged(double x, double y) {
        /* Record where the mouse is. */
        mouse = new Point2D.Double(x, y);
    }
}
