package particles.scenes;
import particles.*;
import java.awt.Color;
import java.awt.image.*;
import javax.imageio.*;
import java.util.*;
import java.awt.geom.*;
import java.io.*;

public class PhotoExploder extends Scene {
    private ParticleSystem system = new ParticleSystem();
    
    private Queue<String> files; // Image files to process

    private int countdown;  // Number of ticks before the animation starts
    private int stepsLeft;  // Steps before we're done.
    private int imagePause; // Number of ticks the image remains for

    /* Countdown timer - in frames. */
    private static final int kCountdownTime = 25;

    /* Number of steps to put the photo back together. */
    private static final int kNumSteps = 40;

    /* Number of steps to wait on the photo for. */
    private static final int kImageSteps = 50;

    /* Background color. */
    private static final Color kBackgroundColor = new Color(0x43, 0x44, 0x50);

    /* Fraction of space dedicated to a border. */
    private static final double kPadding = 0.3;

    /* Speed range. */
    private static final double kMinSpeed = 1;
    private static final double kMaxSpeed = 4;

    /* Pixel spacing. */
    private static final double kPixelSpacing = 3.5;

    /* Gets all image files in a random order. */
    private Queue<String> imageFilesIn(String path) {
        var directory = new File(path);
        if (!directory.exists() || !directory.isDirectory()) {
            throw new RuntimeException(path + " is not a directory.");
        }
    
        var images = new TreeSet<String>();

        /* Find all image files. */
        for (var entry: directory.listFiles()) {
            var file = entry.getName().toLowerCase();
            if (file.endsWith(".png") ||
                file.endsWith(".jpg") ||
                file.endsWith(".jpeg") ||
                file.endsWith(".bmp") ||
                file.endsWith(".gif")) {
                images.add(path + "/" + entry.getName());
            }
        }

        /* Scramble the order. */
        Queue<String> result = new LinkedList<>();
        while (!images.isEmpty()) {
            var next = images.iterator().next();
            result.add(next);
            images.remove(next);
        }
        return result;
    }

    public PhotoExploder() {
        files = imageFilesIn("photos");
        loadNextImage();
    }
    
     /* Fit bounds to aspect ratio. */
    private Rectangle2D.Double fitToBounds(Rectangle2D.Double bounds, double aspectRatio) {
        /* If the rectangle has negative size, clip to 0x0. */
        if (bounds.width <= 0 || bounds.height <= 0) {
            return new Rectangle2D.Double(bounds.x, bounds.y, 0, 0);
        }

        double width, height;

        /* Too narrow? */
        if (bounds.width / bounds.height <= aspectRatio) {
            width = bounds.width;
            height = width / aspectRatio;
        } else {
            height = bounds.height;
            width = height * aspectRatio;
        }

        double baseX = bounds.x + (bounds.width  - width)  / 2.0;
        double baseY = bounds.y + (bounds.height - height) / 2.0;

        return new Rectangle2D.Double(baseX, baseY, width, height);
    }

    private void setImage(String filename) {
        /* Load the image. */
        BufferedImage image;
        try {
            image = ImageIO.read(new File(filename));
        } catch (IOException e) {
            throw new RuntimeException("Error loading image " + filename, e);
        }

        /* Content area - where the image is allowed to go. */
        var contentArea = new Rectangle2D.Double(
            kPadding * SCENE_WIDTH,
            kPadding * SCENE_HEIGHT,
            (1.0 - 2 * kPadding) * SCENE_WIDTH,
            (1.0 - 2 * kPadding) * SCENE_HEIGHT
        );

        /* Get a bounding box for the image. */
        var bounds = fitToBounds(contentArea, (double) image.getWidth() / image.getHeight());

        /* For each pixel in the bounds, create a particle. */
        for (double x = 0; x < bounds.width; x += kPixelSpacing) {
            for (double y = 0; y < bounds.height; y += kPixelSpacing) {
                /* Which pixel should we extract? */
                double srcX = x * image.getWidth()  / bounds.width;
                double srcY = y * image.getHeight() / bounds.height;

                addParticle(bounds.x + x, bounds.y + y, new Color(image.getRGB((int) srcX, (int) srcY)));
            }
        }

        /* Reset the countdown. */
        countdown  = kCountdownTime;
        stepsLeft  = kNumSteps;
        imagePause = kImageSteps;
    }

    private void loadNextImage() {
        /* Loop through images by cycling the queue. */
        var next = files.remove();
        setImage(next);
        files.add(next);
    }

    private void addParticle(double x, double y, Color color) {
        Particle particle = new Particle();

        /* Choose a random angle and speed. */
        double theta = RandomGenerator.randomReal(0, 2 * Math.PI);
        double speed = RandomGenerator.randomReal(kMinSpeed, kMaxSpeed);

        particle.dx = speed * Math.cos(theta);
        particle.dy = speed * Math.sin(theta);

        /* Figure out where this particle will be in a certain number of time steps. */
        particle.x = x - particle.dx * kNumSteps;
        particle.y = y - particle.dy * kNumSteps;

        particle.lifetime = kNumSteps;

        particle.color = color;
        system.add(particle);
    }

    @Override public void tick() {
        /* If the countdown is still active, do nothing. */
        if (countdown > 0) {
            countdown--;
        }
        /* Otherwise, if we're in the active window, move particles. */
        else if (stepsLeft > 0) {
            system.moveParticles();
            stepsLeft--;
        }
        /* Otherwise, if we're at the pause window, marvel. */
        else if (imagePause > 0) {
            imagePause--;
        }
        /* Otherwise, reset the image. */
        else {
            system.moveParticles(); // Clear all particles
            loadNextImage();
        }
    }

    @Override public void draw(ParticleGraphics g) {
        g.setColor(kBackgroundColor);
        g.fillRect(0, 0, SCENE_WIDTH, SCENE_HEIGHT);
        system.drawParticles(g);
    }
}
